﻿using CMS;
using CMS.DataEngine;
using CMS.Ecommerce;
using CMS.EventLog;
using CMS.Globalization;
using CMS.Helpers;
using CMS.Membership;
using CMS.Modules;
using CustomFedEx.FedExRateService;
using System;
using System.Collections.Generic;
using System.Linq;

[assembly: RegisterModule(typeof(CustomFedEx.CustomFedExCarrier))]

namespace CustomFedEx
{
    public class CustomFedExCarrier : ICarrierProvider
    {
        private const string NAME = "Custom FedEx provider";
        private List<KeyValuePair<string, string>> services;

        /// <summary>
        /// Gets the name of the carrier provider.
        /// </summary>
        public string CarrierProviderName
        {
            get
            {
                return NAME;
            }
        }

        /// <summary>
        /// Returns a list of service code names and display names.US
        /// </summary>
        public List<KeyValuePair<string, string>> GetServices()
        {
            return services ?? (services = new List<KeyValuePair<string, string>>
            {
                new KeyValuePair<string, string>("PRIORITY_OVERNIGHT","FedEx Priority Overnight"),
                new KeyValuePair<string, string>("STANDARD_OVERNIGHT", "FedEx Standard Overnight"),
                new KeyValuePair<string, string>("2_DAY","FedEx 2 Day")
            });
        }

        /// <summary>
        /// Calculates the shipping cost of the given delivery.
        /// </summary>
        /// <param name="delivery">The shipped delivery.</param>
        /// <param name="currencyCode">The code of the currency in which the shipping price should be calculated (e.g. USD or EUR).</param>
        /// <returns>The price in the currency specified by currencyCode.</returns>
        public decimal GetPrice(Delivery delivery, string currencyCode)
        {
            decimal decRate = SettingsKeyInfoProvider.GetDecimalValue("FedExDefaultShippingRate");
            try
            {
                ServiceType stType = new ServiceType();
                switch (delivery.ShippingOption.ShippingOptionName)
                {
                    case "FedExPriorityOvernight":
                        stType = ServiceType.PRIORITY_OVERNIGHT;
                        break;
                    case "FedExStandardOvernight":
                        stType = ServiceType.STANDARD_OVERNIGHT;
                        break;
                    case "FedEx2Day":
                        stType = ServiceType.FEDEX_2_DAY;
                        break;
                }

                CurrentUserInfo uinfo = MembershipContext.AuthenticatedUser;

                RateReply reply = new RateReply();
                // Cache the data for 60 minutes with a key
                using (CachedSection<RateReply> cs = new CachedSection<RateReply>(ref reply, 60, true, null, "FedExRatesAPI-" + stType.ToString().Replace(" ", "-") + "-" + uinfo.UserID + "-" + delivery.DeliveryAddress.AddressZip + "-" + ValidationHelper.GetString(delivery.Weight, "")))
                {
                    if (cs.LoadData)
                    {
                        //Create the request
                        RateRequest request = CreateRateRequest(delivery, stType, uinfo);
                        //Create the service
                        RateService service = new RateService();
                        // Set the service url
                        service.Url = SettingsKeyInfoProvider.GetValue("FedExServiceUrl");
                        // Call the web service passing in a RateRequest and returning a RateReply
                        reply = service.getRates(request);
                        // Cache the reply
                        cs.Data = reply;
                    }
                    reply = cs.Data;
                }

                if (reply.HighestSeverity == NotificationSeverityType.SUCCESS)
                {
                    foreach (RateReplyDetail repDetail in reply.RateReplyDetails)
                    {
                        foreach (RatedShipmentDetail rsd in repDetail.RatedShipmentDetails)
                        {
                            decRate = rsd.ShipmentRateDetail.TotalNetFedExCharge.Amount;
                        }
                    }
                }
                else
                {
                    //Clean up the cached value so the next time the value is pulled again
                    CacheHelper.ClearCache("FedExRatesAPI-" + stType.ToString().Replace(" ", "-") + "-" + uinfo.UserID + "-" + delivery.DeliveryAddress.AddressZip + "-" + ValidationHelper.GetString(delivery.Weight, ""));

                    //Log the error
                    if (reply.Notifications != null)
                    {
                        if (reply.Notifications.Count() > 0)
                        {
                            EventLogProvider.LogEvent("ERROR", "FedExCarrier - GetPrice", reply.Notifications[0].Message);
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                //Log the error
                EventLogProvider.LogException("FedExCarrier - GetPrice", "EXCEPTION", ex);
            }

            return decRate;
        }

        /// <summary>
        /// Checks whether the shipping option is available for the specified shipping address.
        /// The only supported shipping addresses are to USA, Canada or Mexico.
        /// </summary>
        public bool CanDeliver(Delivery delivery)
        {
            return true;
        }

        /// <summary>
        /// Returns the Guid for the configuration page. This allows the Store Settings UI to display the configuration page on the carrier details.
        /// </summary>
        public Guid GetConfigurationUIElementGUID()
        {
            UIElementInfo ui = UIElementInfoProvider.GetUIElementInfo("CustomFedExCarrier", "CustomFedExCarrierConfiguration");
            return ui.ElementGUID;
        }

        public Guid GetServiceConfigurationUIElementGUID(string serviceName)
        {
            return Guid.Empty;
        }

        /// <summary>
        /// Returns true if the carrier provider can handle the specified shipping service.
        /// </summary>
        private bool SupportsService(string serviceName)
        {
            return GetServices().Any(item => item.Key == serviceName);
        }


        #region FedEx API Calls

        private static RateRequest CreateRateRequest(Delivery delivery, ServiceType stType, CurrentUserInfo uinfo)
        {
            // Build a RateRequest
            RateRequest request = new RateRequest();
            //
            request.WebAuthenticationDetail = new WebAuthenticationDetail();
            request.WebAuthenticationDetail.UserCredential = new WebAuthenticationCredential();
            request.WebAuthenticationDetail.UserCredential.Key = SettingsKeyInfoProvider.GetValue("FedExUserCredentialKey");
            request.WebAuthenticationDetail.UserCredential.Password = SettingsKeyInfoProvider.GetValue("FedExUserCredentialPassword");
            request.WebAuthenticationDetail.ParentCredential = new WebAuthenticationCredential();
            request.WebAuthenticationDetail.ParentCredential.Key = SettingsKeyInfoProvider.GetValue("FedExUserCredentialKey");
            request.WebAuthenticationDetail.ParentCredential.Password = SettingsKeyInfoProvider.GetValue("FedExUserCredentialPassword");
            //
            request.ClientDetail = new ClientDetail();
            request.ClientDetail.AccountNumber = SettingsKeyInfoProvider.GetValue("FedExAccountNumber");
            request.ClientDetail.MeterNumber = SettingsKeyInfoProvider.GetValue("FedExMeterNumber");
            //
            request.TransactionDetail = new TransactionDetail();
            request.TransactionDetail.CustomerTransactionId = "***Rate Request using VC#***"; // This is a reference field for the customer.  Any value can be used and will be provided in the response.
            //
            request.Version = new VersionId();
            //
            request.ReturnTransitAndCommit = true;
            request.ReturnTransitAndCommitSpecified = true;
            //
            SetShipmentDetails(request, delivery, stType);
            //
            return request;
        }

        private static void SetShipmentDetails(RateRequest request, Delivery delivery, ServiceType stType)
        {
            request.RequestedShipment = new RequestedShipment();
            request.RequestedShipment.ShipTimestamp = DateTime.Now; // Shipping date and time
            request.RequestedShipment.ShipTimestampSpecified = true;
            request.RequestedShipment.DropoffType = DropoffType.REGULAR_PICKUP; //Drop off types are BUSINESS_SERVICE_CENTER, DROP_BOX, REGULAR_PICKUP, REQUEST_COURIER, STATION
            request.RequestedShipment.ServiceType = stType; // Service types are STANDARD_OVERNIGHT, PRIORITY_OVERNIGHT, FEDEX_GROUND ...
            request.RequestedShipment.ServiceTypeSpecified = true;
            request.RequestedShipment.PackagingType = PackagingType.FEDEX_PAK; // Packaging type FEDEX_BOX, FEDEX_PAK, FEDEX_TUBE, YOUR_PACKAGING, ...
            request.RequestedShipment.PackagingTypeSpecified = true;
            //
            SetOrigin(request, delivery);
            //
            SetDestination(request, delivery);
            //
            SetPackageLineItems(request, delivery);
            //
            request.RequestedShipment.PackageCount = ValidationHelper.GetString(request.RequestedShipment.RequestedPackageLineItems.Length, "1");
        }

        private static void SetOrigin(RateRequest request, Delivery delivery)
        {
            request.RequestedShipment.Shipper = new Party();
            request.RequestedShipment.Shipper.Address = new Address();
            request.RequestedShipment.Shipper.Address.StreetLines = new string[1] { SettingsKeyInfoProvider.GetValue("FedExOriginStreet") };
            request.RequestedShipment.Shipper.Address.City = SettingsKeyInfoProvider.GetValue("FedExOriginCity");
            request.RequestedShipment.Shipper.Address.StateOrProvinceCode = SettingsKeyInfoProvider.GetValue("FedExOriginState");
            request.RequestedShipment.Shipper.Address.CountryCode = SettingsKeyInfoProvider.GetValue("FedExOriginCountry");
            request.RequestedShipment.Shipper.Address.PostalCode = SettingsKeyInfoProvider.GetValue("FedExOriginPostalCode");
        }

        private static void SetDestination(RateRequest request, Delivery delivery)
        {
            request.RequestedShipment.Recipient = new Party();
            request.RequestedShipment.Recipient.Address = new Address();
            request.RequestedShipment.Recipient.Address.StreetLines = new string[1] { delivery.DeliveryAddress.AddressLine1 };
            request.RequestedShipment.Recipient.Address.City = delivery.DeliveryAddress.AddressCity;
            request.RequestedShipment.Recipient.Address.PostalCode = delivery.DeliveryAddress.AddressZip;
            request.RequestedShipment.Recipient.Address.CountryCode = CountryInfoProvider.GetCountryInfo(delivery.DeliveryAddress.AddressCountryID).CountryTwoLetterCode;
            request.RequestedShipment.Recipient.Address.StateOrProvinceCode = StateInfoProvider.GetStateInfo(delivery.DeliveryAddress.AddressStateID).StateCode;
        }

        private static void SetPackageLineItems(RateRequest request, Delivery delivery)
        {
            int i = 1;
            List<RequestedPackageLineItem> lstItems = new List<RequestedPackageLineItem>();
            foreach (DeliveryItem item in delivery.Items)
            {
                SKUInfo sku = item.Product;
                RequestedPackageLineItem ritem = new RequestedPackageLineItem();
                ritem.SequenceNumber = i.ToString(); // package sequence number
                ritem.GroupPackageCount = i.ToString();
                // package weight
                ritem.Weight = new Weight();
                ritem.Weight.Units = WeightUnits.LB;
                ritem.Weight.UnitsSpecified = true;
                ritem.Weight.Value = ValidationHelper.GetDecimal(sku.SKUWeight, 1) * item.Amount;
                ritem.Weight.ValueSpecified = true;
                // package dimensions
                ritem.Dimensions = new Dimensions();
                ritem.Dimensions.UnitsSpecified = false;
                lstItems.Add(ritem);
                i += 1;
            }
            request.RequestedShipment.RequestedPackageLineItems = lstItems.ToArray();
        }

        #endregion

    }
}